/**
This program reads an IDX file, extract a 2D slice from it, group the samples in
groups of 4x4 and write the data as a raw file.
*/

#include <idx/idx_file.h>
#include <idx/idx.h>
#include <algorithm>
#include <iostream>
#include <fstream>

using namespace std;
using namespace hana;
using namespace hana::core;
using namespace hana::idx;

int main()
{
    /* Read a slice from an idx file into memory */
    IdxFile idx_file;

    idx::Error error = read_idx_file("D:/Datasets/miranda/density-384-384-256.idx", &idx_file);
    if (error.code != core::Error::NoError) {
        cout << "Error: " << error.get_error_msg() << "\n";
        return EXIT_FAILURE;
    }

    int hz_level = idx_file.get_max_hz_level();
    int field = idx_file.get_field_index("data");
    int time = idx_file.get_min_time_step();

    Grid grid;
    grid.extent.from = Vector3i(96, 96, 127);
    grid.extent.to = Vector3i(111, 111, 127);
    grid.data.bytes = idx_file.get_size_inclusive(grid.extent, field, hz_level);
    grid.data.ptr = (char*)malloc(grid.data.bytes);

    Vector3i from, to, stride;
    idx_file.get_grid_inclusive(grid.extent, hz_level, from, to, stride);
    Vector3i dim = (to - from) / stride + 1;
    cout << "Resulting grid dim = " << dim.x << " x " << dim.y << " x " << dim.z << "\n";

    error = read_idx_grid_inclusive(idx_file, field, time, hz_level, &grid);
    idx::deallocate_memory();

    if (error.code != core::Error::NoError) {
        cout << "Error: " << error.get_error_msg() << "\n";
        return EXIT_FAILURE;
    }

    /* Group the samples into groups of 4x4, assuming the data is in row-major order */
    if ((dim.x > 1 && dim.x % 4 != 0) ||
        (dim.y > 1 && dim.y % 4 != 0) ||
        (dim.z > 1 && dim.z % 4 != 0)) {
        cout << "Error: dimension not multiple of 4\n";
        return EXIT_FAILURE;
    }

    size_t output_size = grid.data.bytes;
    double* output = (double*)malloc(output_size);
    double* input = (double*)(grid.data.ptr);
    uint64_t sx = dim.x;
    uint64_t sxy = dim.x * dim.y;
    int brick_dim_x = min(4, dim.x);
    int brick_dim_y = min(4, dim.y);
    int brick_dim_z = min(4, dim.z);
    uint64_t dx = brick_dim_x;
    uint64_t dxy = dx * brick_dim_y;
    uint64_t dxyz = dxy * brick_dim_z;
    for (int z = 0; z < dim.z; z += 4) {
        for (int y = 0; y < dim.y; y += 4) {
            for (int x = 0; x < dim.x; x += 4) {
                uint64_t out_start = ((z / 4) * (dim.y / 4) * (dim.x / 4) +
                                      (y / 4) * (dim.x / 4) + (x / 4)) * dxyz;
                uint64_t in_start = x + y * sx + z * sxy;
                for (int k = 0; k < brick_dim_z; ++k) {
                    for (int j = 0; j < brick_dim_y; ++j) {
                        for (int i = 0; i < brick_dim_x; ++i) {
                            uint64_t ijk = i + j * dx + k * dxy;
                            uint64_t xyz = i + j * sx + k * sxy;
                            output[out_start + ijk] = input[in_start + xyz];
                        }
                    }
                }
            }
        }
    }

    FILE* fp = fopen("data.raw", "wb");
    fwrite(grid.data.ptr, grid.data.bytes, 1, fp);
    fclose(fp);

    fp = fopen("data-brick.raw", "wb");
    fwrite(output, output_size, 1, fp);
    fclose(fp);

    free(grid.data.ptr);
    free(output);

    return EXIT_SUCCESS;
}
