/**
This program reads low resolution 2D data from a bricked IDX file, and upsamples
the data using bi-Laplacian upsampling, and write the upsampled data out as a
.raw file.
*/

#include <idx/idx_file.h>
#include <idx/idx.h>
#include <Eigen/SparseCore>
#include <Eigen/IterativeLinearSolvers>
#include <Eigen/SparseQR>
#include <Eigen/SparseLU>
#include <Eigen/SVD>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <vector>

using namespace std;
using namespace hana;
using namespace hana::core;
using namespace hana::idx;

// Return a sample's index in the list of unknown samples, given its x and y
// We assume that a sample is known if ((y / 4 - from.y / 4) % (stride.y / 4) == 0)
// NOTE: for now we assume from = (0, 0, 0)
inline int get_sample_index(const Vector3i& from, const Vector3i& stride,
                            int num_cols, int x, int y)
{
    // we need to know how many unknown samples are there in a row
    int nc = num_cols;
    int num_unknowns_per_row = nc - ((((nc-1)/4) / (stride.x/4)) * 4 + min((nc-1)%stride.x+1, 4));;
    // how many rows are there that contain both known and unknown values
    int num_unknown_rows = 0;
    if (y > 0) {
        num_unknown_rows =(((y-1)/4) / (stride.y/4)) * 4 + min((y-1)%stride.y+1, 4);
    }
    // how many unknown samples are there in the last row?
    int num_unknown_last_row = x + 1;
    if ((y/4) % (stride.y/4) == 0) {
        num_unknown_last_row = (x+1) - (((x/4) / (stride.x/4)) * 4 + min(x%stride.x+1, 4));
    }
    return num_unknowns_per_row * num_unknown_rows +
           (y - num_unknown_rows) * num_cols + num_unknown_last_row - 1;
}

int main()
{
    /* Read a slice from an idx file into memory. NOTE: the slice is at some low
    resolution at each sample is a group of (4x4) original samples. */
    IdxFile idx_file;

    //idx::Error error = read_idx_file("D:/Workspace/Interpolation/Data/miranda_slice_brick/data-brick.idx", &idx_file);
    idx::Error error = read_idx_file("D:/Workspace/Interpolation/Data/sine_16x16/sine-brick.idx", &idx_file);
    //idx::Error error = read_idx_file("D:/Workspace/Interpolation/Data/small_16x16/small_brick.idx", &idx_file);
    if (error.code != core::Error::NoError) {
        cout << "Error: " << error.get_error_msg() << "\n";
        return EXIT_FAILURE;
    }

    int hz_level = idx_file.get_max_hz_level() - 2;
    int field = idx_file.get_field_index("data");
    int time = idx_file.get_min_time_step();

    Grid grid;
    grid.extent.from = Vector3i(0, 0, 0);
    grid.extent.to = Vector3i(3, 3, 0);
    //grid.extent.to = Vector3i(95, 95, 0);
    grid.data.bytes = idx_file.get_size_inclusive(grid.extent, field, hz_level);
    grid.data.ptr = (char*)malloc(grid.data.bytes);

    Vector3i from, to, stride;
    idx_file.get_grid_inclusive(grid.extent, hz_level, from, to, stride);
    Vector3i dims = (to - from) / stride + 1;
    cout << "Resulting grid dim = " << dims.x << " x " << dims.y << " x " << dims.z << "\n";

    error = read_idx_grid_inclusive(idx_file, field, time, hz_level, &grid);
    idx::deallocate_memory();

    if (error.code != core::Error::NoError) {
        cout << "Error: " << error.get_error_msg() << "\n";
        return EXIT_FAILURE;
    }

    /* Put the data back to a larger (original) grid */
    int sample_size = idx_file.get_field_sample_size(field) / 16;
    Vector3i original_dims = (grid.extent.to - grid.extent.from + 1) * 4;
    uint64_t original_grid_size = original_dims.x * original_dims.y * sample_size;
    double* original_grid = (double*)calloc(original_grid_size, sizeof(char));
    double* input = (double*)grid.data.ptr;
    from = from * 4;
    to = to * 4;
    stride = stride * 4;
    for (int y = from.y; y <= to.y; y += stride.y) {
        int yy = (y - from.y) / stride.y;
        for (int x = from.x; x <= to.x; x += stride.x) {
            int xx = (x - from.x) / stride.x;
            uint64_t out_start = x + y * original_dims.x;
            uint64_t in_start = (xx + yy * dims.x) * 4 * 4;
            for (int j = 0; j < 4; ++j) {
                for (int i = 0; i < 4; ++i) {
                    uint64_t xy = i + j * original_dims.x;
                    uint64_t ij = i + j * 4;
                    original_grid[out_start + xy] = input[in_start + ij];
                }
            }
        }
    }

    /* Set up the matrix (A) and the right hand side (b) */
    // loop through each element in the data, find out where that element is in
    // the upsampled array
    using namespace Eigen;
    double weights_1[] = { 1, -4, 6, -4, 1 }; // one dimensional weights
    double weights_2[][5] = { { 11, -56, 114, -104, 35 },
                              { -1, 4, 6, -20, 11 },
                              { -1, 16, -30, 16, -1 },
                              { 11, -20, 6, 4, -1 },
                              { 35, -104, 114, -56, 11 } };
    double weights_3[] = { 1, -2, 1 }; // laplacian
    for (auto& r : weights_2) {
        for (auto& v : r) {
            v /= 12.0;
        }
    }
    int num_known_vals = dims.x * dims.y * 4 * 4;
    int M = original_dims.x * original_dims.y;
    int N = M - num_known_vals;
    int unknown_dim_x = original_dims.x - dims.x * 4;
    int unknown_dim_y = original_dims.y - dims.y * 4;
    SparseMatrix<double> A(N, N);
    VectorXd X0(N), B(N);
    X0.setZero();
    B.setZero();
    vector<Triplet<double>> triplet_list;
    int num_known = 0;
    for (int y = 0; y < original_dims.y; ++y) {
        bool known_val_y = (y / 4 - from.y / 4) % (stride.y / 4) == 0;
        int start_mask_y = -2; // mask shifts = { -2, -1, 0, 1, 2 }
        if (y < 1) {
            start_mask_y = 0; // mask shifts = { 0, 1, 2, 3, 4 }
        }
        else if (y < 2) {
            start_mask_y = -1; // mask shifts = { -1, 0, 1, 2, 3 }
        }
        else if (y + 1 > original_dims.y - 1) {
            start_mask_y = -4; // mask shifts = { -4, -3, -2, -1, 0 }
        }
        else if (y + 2 > original_dims.y - 1) {
            start_mask_y = -3; // mask shifts = { -3, -2, -1, 0, 1 }
        }
        for (int x = 0; x < original_dims.x; ++x) {
            bool known_val_x = (x / 4 - from.x / 4) % (stride.x / 4) == 0;
            if (known_val_x && known_val_y) {
                continue;
            }
            int urow, ucol;
            urow = ucol = get_sample_index(from, stride, original_dims.x, x, y);
            //cout << x << " " << y << " " << urow << "\n";
            int col = x + y * original_dims.x;
            // set the bi-laplacian to 0
            int start_mask_x = -2; // mask shifts = { -2, -1, 0, 1, 2 }
            if (x < 1) {
                start_mask_x = 0; // mask shifts = { 0, 1, 2, 3, 4 }
            }
            else if (x < 2) {
                start_mask_x = -1; // mask shifts = { -1, 0, 1, 2, 3 }
            }
            else if (x + 1 > original_dims.x - 1) {
                start_mask_x = -4; // mask shifts = { -4, -3, -2, -1, 0 }
            }
            else if (x + 2 > original_dims.x - 1) {
                start_mask_x = -3; // mask shifts = { -3, -2, -1, 0, 1 }
            }

            // weights in x
            for (int i = 0; i < 5; ++i) {
                int xx = x + start_mask_x + i;
                int c = xx + y * original_dims.x;
                if ((y / 4 - from.y / 4) % (stride.y / 4) == 0 &&
                    (xx / 4 - from.x / 4) % (stride.x / 4) == 0) { // known
                    B[urow] -= original_grid[c] * weights_1[i];
                }
                else {
                    int uc = get_sample_index(from, stride, original_dims.x, xx, y);
                    HANA_ASSERT(urow >= 0 && uc >= 0);
                    triplet_list.push_back(Triplet<double>(urow, uc, weights_1[i]));
                }
            }

            // weights in y
            for (int j = 0; j < 5; ++j) {
                int yy = y + start_mask_y + j;
                int c = x + yy * original_dims.x;
                if ((x / 4 - from.x / 4) % (stride.x / 4) == 0 &&
                    (yy / 4 - from.y / 4) % (stride.y / 4) == 0) { // known
                    B[urow] -= original_grid[c] * weights_1[j];
                }
                else {
                    int uc = get_sample_index(from, stride, original_dims.x, x, yy);
                    HANA_ASSERT(urow >= 0 && uc >= 0);
                    triplet_list.push_back(Triplet<double>(urow, uc, weights_1[j]));
                }
            }

            // weights in xy
            const auto& weights_2_y = weights_2[start_mask_y + 4];
            const auto& weights_2_x = weights_2[start_mask_x + 4];
            for (int j = 0; j < 5; ++j) {
                int yy = y + start_mask_y + j;
                for (int i = 0; i < 5; ++i) {
                    int xx= x + start_mask_x + i;
                    int c = xx + yy * original_dims.x;
                    auto weight = 2 * weights_2_y[j] * weights_2_x[i];
                    if ((yy / 4 - from.y / 4) % (stride.y / 4) == 0 &&
                        (xx / 4 - from.x / 4) % (stride.x / 4) == 0) {
                        B[urow] -= original_grid[c] * weight;
                    }
                    else {
                        int uc = get_sample_index(from, stride, original_dims.x, xx, yy);
                        HANA_ASSERT(urow >= 0 && uc >= 0);
                        triplet_list.push_back(Triplet<double>(urow, uc, weight));
                    }
                }
            }
        }
    }
    /*ofstream X0_stream("X0.txt");
    ofstream B_stream("B.txt");
    X0_stream << "---------------------------------------\n" << X0 << "\n";
    B_stream << "---------------------------------------\n" << B << "\n";*/
    cout << "Num known values = " << num_known << "\n";
    cout << "Num known values should be = " << dims.x * dims.y * 4 * 4 << "\n";

    /* Solve the linear system */
    A.setFromTriplets(triplet_list.begin(), triplet_list.end());
    /*A.makeCompressed();
    SparseQR<SparseMatrix<double>, NaturalOrdering<int>> QR(A);
    cout << "rank = " << QR.rank() << " N = " << N << "\n";*/
    //return 0;

    BiCGSTAB<SparseMatrix<double>, IncompleteLUT<double>> solver;
    SparseLU<SparseMatrix<double>, COLAMDOrdering<int>> solver2;
    solver2.analyzePattern(A);
    solver2.factorize(A);
    //BiCGSTAB<SparseMatrix<double>> solver;
    // Comment out the following two lines for high hz levels
    solver.preconditioner().setFillfactor(100); // need this for convergence
    solver.preconditioner().setDroptol(1e-3); // set to 3 for higher hz, 9 for lower hz
    /*for (int i = 0; i < N; ++i) {
        double b = B[i];
        if (b != b) {
            cout << "Fucking nan\n";
        }
        else if (b != 0) {
            cout << b << " ";
        }
    }*/

    //MatrixXd C(N, N);
    //C.setZero();
    //vector<vector<double>> Z;
    //Z.resize(N);
    //for (auto & r : Z) {
    //    r.resize(N, 0);
    //}

    //for (int k = 0; k < A.outerSize(); ++k) {
    //    for (SparseMatrix<double>::InnerIterator it(A, k); it; ++it)
    //    {
    //        //cout << it.col() << " " << it.row() << " " << it.value() << ", ";
    //        Z[it.row()][it.col()] = it.value();
    //        C(it.row(), it.col()) = it.value();
    //    }
    //    //cout << endl;
    //}

    /*for (auto & r : Z) {
        for (auto v : r) {
            if (v != 0)
                cout << v << " ";
            else
                cout << "  ";
        }
        cout << "\n";
    }*/
    //cout << C << "\n";
    /*JacobiSVD<MatrixXd> svd(C, ComputeThinU | ComputeThinV);
    cout << "singular values = " << svd.singularValues() << endl;*/

    solver.compute(A);
    VectorXd X(N);
    solver.setTolerance(1e-3); // this can be set to 1e-9 for high hz levels
    cout << solver.tolerance() << "\n";
    //X = solver.solveWithGuess(B, X0);
    X = solver2.solve(B);
    cout << "Iteration = " << solver.iterations() << "\n";
    cout << "Error = " << solver.error() << "\n";

    /* Put back the result */
    double* output = (double*)malloc(original_grid_size);
    for (int y = 0; y < original_dims.y; ++y) {
        bool known_y = (y / 4 - from.y / 4) % (stride.y / 4) == 0;
        for (int x = 0; x < original_dims.x; ++x) {
            bool known_x = (x / 4 - from.x / 4) % (stride.x / 4) == 0;
            if (known_x && known_y) {
                output[x + y * original_dims.x] = original_grid[x + y * original_dims.x];
            }
            else {
                int idx = get_sample_index(from, stride, original_dims.x, x, y);
                output[x + y * original_dims.x] = X[idx];

            }
        }
    }

    /* Compute RMS error */
    double* ground_truth = (double*)malloc(original_grid_size);
    //FILE* fp = fopen("D:/Workspace/Interpolation/Data/small_16x16/data.raw", "rb");
    //FILE* fp = fopen("D:/Workspace/Interpolation/Data/miranda_slice_brick/data.raw", "rb");
    FILE* fp = fopen("D:/Workspace/Interpolation/Data/sine_16x16/data.raw", "rb");
    fread(ground_truth, original_grid_size, 1, fp);
    ofstream reconstruct("reconstruct.txt");
    ofstream original("original.txt");
    double err = 0;
    for (int y = 0; y < original_dims.y; ++y) {
        for (int x = 0; x < original_dims.x; ++x) {
            int idx = x + y * original_dims.x;
            auto diff = fabs(output[idx] - ground_truth[idx]);
            err += diff * diff;
            reconstruct << output[idx] << "\n";
            original << ground_truth[idx] << "\n";
        }
    }
    cout << "RMS = " << sqrt(err / (original_dims.x * original_dims.y)) << "\n";


    /* test the bi-laplacian conditions */
    vector<double> bilap(original_dims.x * original_dims.y, 0);
    for (int y = 0; y < original_dims.y; ++y) {
        int start_mask_y = -2; // mask shifts = { -2, -1, 0, 1, 2 }
        if (y < 1) {
            start_mask_y = 0; // mask shifts = { 0, 1, 2, 3, 4 }
        }
        else if (y < 2) {
            start_mask_y = -1; // mask shifts = { -1, 0, 1, 2, 3 }
        }
        else if (y + 1 > original_dims.y - 1) {
            start_mask_y = -4; // mask shifts = { -4, -3, -2, -1, 0 }
        }
        else if (y + 2 > original_dims.y - 1) {
            start_mask_y = -3; // mask shifts = { -3, -2, -1, 0, 1 }
        }
        for (int x = 0; x < original_dims.x; ++x) {
            // set the bi-laplacian to 0
            int start_mask_x = -2; // mask shifts = { -2, -1, 0, 1, 2 }
            if (x < 1) {
                start_mask_x = 0; // mask shifts = { 0, 1, 2, 3, 4 }
            }
            else if (x < 2) {
                start_mask_x = -1; // mask shifts = { -1, 0, 1, 2, 3 }
            }
            else if (x + 1 > original_dims.x - 1) {
                start_mask_x = -4; // mask shifts = { -4, -3, -2, -1, 0 }
            }
            else if (x + 2 > original_dims.x - 1) {
                start_mask_x = -3; // mask shifts = { -3, -2, -1, 0, 1 }
            }
            // weights in x
            for (int i = 0; i < 5; ++i) {
                int xx = x + start_mask_x + i;
                int c = xx + y * original_dims.x;
                bilap[c] += output[c] * /*weights_1[i]*/ weights_3[i];
            }
            // weights in y
            for (int j = 0; j < 5; ++j) {
                int yy = y + start_mask_y + j;
                int c = x + yy * original_dims.x;
                bilap[c] += output[c] * /*weights_1[j]*/ weights_3[j];
            }
            // weights in xy
            const auto& weights_2_y = weights_2[start_mask_y + 4];
            const auto& weights_2_x = weights_2[start_mask_x + 4];
            for (int j = 0; j < 5; ++j) {
                int yy = y + start_mask_y + j;
                for (int i = 0; i < 5; ++i) {
                    int xx = x + start_mask_x + i;
                    auto weight = 2 * weights_2_y[j] * weights_2_x[i];
                    int c = xx + yy * original_dims.x;
                    bilap[c] += output[c] * weight;
                }
            }
        }
    }

    ofstream bilap_file("bilap.raw", ios::binary);
    bilap_file.write((const char*)&bilap[0], bilap.size() * sizeof(double));
    /*ofstream bilap_text_stream("bilap.txt");
    for (int i = 0; i < original_dims.y; ++i) {
        for (int j = 0; j < original_dims.x; ++j) {
            bilap_text_stream << bilap[i * original_dims.x + j] << " ";
        }
        bilap_text_stream << "\n";
    }*/


    /*fp = fopen("data.raw", "wb");
    fwrite(original_grid, original_grid_size, 1, fp);
    fclose(fp);*/
    fp = fopen("output.raw", "wb");
    fwrite(output, original_grid_size, 1, fp);
    fclose(fp);

    free(grid.data.ptr);
    free(original_grid);
    free(output);

    return EXIT_SUCCESS;
}